#include watcher.js
#include settings/scoped.js
#include settings/virtual.js
#include settings/webstorage.js

const MamiSettings = function(storageOrPrefix) {
    if(typeof storageOrPrefix === 'string')
        storageOrPrefix = new MamiSettingsWebStorage(window.localStorage, storageOrPrefix);
    else if(typeof storageOrPrefix !== 'object')
        throw 'storageOrPrefix must be a prefix string or an object';

    if(typeof storageOrPrefix.get !== 'function'
        || typeof storageOrPrefix.set !== 'function'
        || typeof storageOrPrefix.delete !== 'function')
        throw 'required methods do not exist in storageOrPrefix object';

    const storage = new MamiSettingsVirtualStorage(storageOrPrefix),
        watchers = new MamiWatchers,
        settings = new Map;

    const getSetting = name => {
        const setting = settings.get(name);
        if(setting === undefined)
            throw `setting ${name} is undefined`;
        return setting;
    };

    const getValue = setting => {
        if(setting.immutable)
            return setting.fallback;

        const value = storage.get(setting.name);
        return value === null ? setting.fallback : value;
    };

    const deleteValue = setting => {
        if(setting.immutable)
            return;

        storage.delete(setting.name);
    };

    const setValue = (setting, value) => {
        if(value !== null) {
            if(value === undefined)
                value = null;
            else if('type' in setting) {
                if(Array.isArray(setting.type)) {
                    if(!setting.type.includes(value))
                        throw `setting ${setting.name} must match an enum value`;
                } else {
                    const type = typeof value;
                    let resolved = false;

                    if(type !== setting.type) {
                        if(type === 'string') {
                            if(setting.type === 'number') {
                                value = parseFloat(value);
                                resolved = true;
                            } else if(setting.type === 'boolean') {
                                value = !!value;
                                resolved = true;
                            }
                        } else if(setting.type === 'string') {
                            value = value.toString();
                            resolved = true;
                        }
                    } else resolved = true;

                    if(!resolved)
                        throw `setting ${setting.name} must be of type ${setting.type}`;
                }
            }
        }

        if(setting.immutable)
            return;

        if(value === null || value === setting.fallback) {
            value = setting.fallback;
            storage.delete(setting.name);
        } else
            storage.set(setting.name, value);

        watchers.call(setting.name, value, setting.name);
    };

    const pub = {
        define: (name, type, fallback, immutable, critical, virtual) => {
            if(typeof name !== 'string')
                throw 'setting name must be a string';
            if(typeof type !== 'string' && !Array.isArray(type))
                throw 'type must be a javascript type or array of valid string values.';
            if(settings.has(name))
                throw `setting ${name} has already been defined`;

            settings.set(name, Object.freeze({
                name: name,
                type: type === null ? undefined : type,
                fallback: fallback === undefined ? null : fallback,
                immutable: immutable === true,
                critical: critical === true,
            }));

            if(virtual === true)
                storage.virtualise(name);

            watchers.define(name);
        },
        info: name => getSetting(name),
        names: () => Array.from(settings.keys()),
        has: name => {
            const setting = settings.get(name);
            return setting !== undefined
                && !setting.immutable
                && storage.get(setting.name) !== null;
        },
        get: name => getValue(getSetting(name)),
        set: (name, value) => setValue(getSetting(name), value),
        delete: name => {
            const setting = getSetting(name);
            if(!setting.immutable)
                storage.delete(setting.name)
        },
        toggle: name => {
            const setting = getSetting(name);
            if(!setting.immutable)
                setValue(setting, !getValue(setting));
        },
        touch: name => {
            const setting = getSetting(name);
            watchers.call(setting.name, getValue(setting), setting.name);
        },
        clear: (criticalOnly, prefix) => {
            for(const setting of settings.values())
                if((prefix === undefined || setting.name.startsWith(prefix)) && (!criticalOnly || setting.critical))
                    storage.delete(setting.name);
        },
        watch: (name, handler) => {
            const setting = getSetting(name);
            watchers.watch(setting.name, handler, getValue(setting), setting.name);
        },
        unwatch: (name, handler) => {
            watchers.unwatch(getSetting(name).name, handler);
        },
        virtualise: name => storage.virtualise(getSetting(name).name),
        scope: name => new MamiSettingsScoped(pub, name),
    };

    return pub;
};
