#include sound/sndlibrary.js

const MamiSoundPack = function(name, isReadOnly, title, events) {
    name = (name || '').toString();
    isReadOnly = !!isReadOnly;
    title = (title || '').toString();
    events = events || new Map;

    return {
        getName: () => name,
        isReadOnly: () => isReadOnly,
        getTitle: () => title,
        setTitle: newTitle => {
            if(isReadOnly)
                throw 'Cannot edit read only sound pack.';
            title = (newTitle || '').toString();
        },
        getEventNames: () => Array.from(events.keys()),
        hasEventSound: eventName => events.has(eventName),
        getEventSound: eventName => {
            if(!events.has(eventName))
                throw 'event not registered';
            return events.get(eventName);
        },
        setEventSound: (eventName, soundName) => {
            events.set(
                (eventName || '').toString(),
                (soundName || '').toString()
            );
        },
        removeEventSound: eventName => {
            events.delete(eventName);
        },
        clearEventSounds: () => {
            events.clear();
        },
        clone: (newName, readOnly) => {
            return new MamiSoundPack(newName, readOnly, 'Clone of ' + title, new Map(events));
        },
    };
};

const MamiSoundPackPlayer = function(sndLibrary) {
    let pack;

    return {
        loadPack: packInfo => {
            if(typeof packInfo !== 'object' || typeof packInfo.getEventSound !== 'function')
                throw 'pack is not a valid soundpack';
            pack = packInfo;
        },
        unloadPack: () => {
            pack = undefined;
            buffers.clear();
        },
        hasPack: () => pack !== undefined,

        hasEvent: name => pack !== undefined && pack.hasEventSound(event),
        playEvent: async event => {
            if(pack === undefined)
                return;

            if(Array.isArray(event)) {
                const names = event;
                event = undefined;

                for(const name of names) {
                    if(pack.hasEventSound(name)) {
                        event = name;
                        break;
                    }
                }

                if(event === undefined)
                    return;
            } else if(!pack.hasEventSound(event))
                return;

            await sndLibrary.play(pack.getEventSound(event));
        },
    };
};

const MamiSoundPacks = function() {
    const packs = new Map;

    const registerPack = (packInfo, readOnly) => {
        if(Array.isArray(packInfo)) {
            for(const info of packInfo)
                registerPack(info, readOnly);
            return;
        }

        if(typeof packInfo !== 'object')
            throw 'packInfo must be an object';
        if(typeof packInfo.name !== 'string')
            throw 'packInfo must contain a name field';
        if(typeof packInfo.events !== 'object')
            throw 'packInfo must contain a events field';
        if(packs.has(packInfo.name))
            throw 'a pack with that name has already been registered';

        const events = new Map;
        for(const eventName in packInfo.events) {
            if(typeof packInfo.events[eventName] !== 'string')
                continue;
            events.set(eventName, packInfo.events[eventName]);
        }

        if(events.size < 1)
            throw 'packInfo does not contain any valid events';

        packInfo = new MamiSoundPack(packInfo.name, readOnly, packInfo.title, events);
        packs.set(packInfo.getName(), packInfo);

        return packInfo;
    };

    const getPack = name => {
        if(!packs.has(name))
            throw 'No pack with this name has been registered.';
        return packs.get(name);
    };

    return {
        register: registerPack,
        unregister: name => {
            packs.delete(name);
        },
        clear: () => {
            packs.clear();
        },
        info: name => getPack(name),
        names: () => Array.from(packs.keys()),
        has: name => packs.has(name),
        get: getPack,
    };
};
